/*
 * Created on Jun 29, 2004
 *
 */
package pupil;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;
import java.lang.System;


/**
 * @author chris
 * 
 * Test class for reading in eyelink data files
 */
public class ReadPupil {

	/** the file name */
	private String file;

	/** the input stream for getting the data */
	private BufferedReader inputStream;

	/** left pupil data packets */
	private Vector dataPacketsL;

	/** right pupil data packets */
	private Vector dataPacketsR;

	/** the left eye */
	public static final int LEFT = 0;

	/** the right eye */
	public static final int RIGHT = 1;

	/** both eyes */
	public static final int ALL_TRIALS = -1;

	public static final int TIME_INCREMENT = 4;

	public static final int FILTER_WINDOW_SIZE = 10;

	public static final double MINIMUM_VALID_SIZE = 11.5;


	/** left pupil start indices */
	private Vector trialStartL;

	/** right pupil start indices */
	private Vector trialStartR;

	/**
	 * 
	 * @param file -
	 *            the data file to read
	 */
	public ReadPupil(String file) {
		this.file = file;
		inputStream = null;
		dataPacketsL = null;
		dataPacketsR = null;
	}

	/**
	 * method that parses the data file
	 * 
	 * @throws IOException
	 * @throws NumberFormatException
	 */
	public void read() throws NumberFormatException, IOException {
		String tmp;
		String[] lineSplit;

		dataPacketsL = new Vector();
		dataPacketsR = new Vector();

		trialStartL = new Vector();
		trialStartR = new Vector();

		boolean hasSamples = hasSampleData();



		try {
			inputStream = new BufferedReader(new FileReader(file));
		} catch (Exception ex) {
			System.out.println(file + " not found!");
			System.exit(1);
		}


		if (! hasSamples) {
			// parse the original way
			while ((tmp = inputStream.readLine()) != null) {

				lineSplit = tmp.split("\\s+");

				//start event
				if (lineSplit[0].equals("START")) {

					//the timestamp of this event
					long timeStamp = Long.parseLong(lineSplit[1]);

					trialStartL.add(new Integer(dataPacketsL.size()));
					dataPacketsL.add(new StartPacket(timeStamp, trialStartL.size()));

					trialStartR.add(new Integer(dataPacketsR.size()));
					dataPacketsR.add(new StartPacket(timeStamp, trialStartR.size()));

				} else if (lineSplit[0].equals("END")) {
					//the timestamp of this event
					long timeStamp = Long.parseLong(lineSplit[1]);

					//sanity checks...
					assert !dataPacketsL.isEmpty() && !dataPacketsR.isEmpty();
					assert (dataPacketsL.isEmpty() || ((DataPacket) dataPacketsL
						.lastElement()).getTimeStamp() <= timeStamp) : ((DataPacket) dataPacketsL
						.lastElement()).getTimeStamp()
						+ "!<=" + timeStamp;
					assert (dataPacketsR.isEmpty() || ((DataPacket) dataPacketsR
						.lastElement()).getTimeStamp() <= timeStamp) : ((DataPacket) dataPacketsR
						.lastElement()).getTimeStamp()
						+ "!<=" + timeStamp;

					dataPacketsL.add(new EndPacket(timeStamp, trialStartL.size()));
					dataPacketsR.add(new EndPacket(timeStamp, trialStartR.size()));

				} else if (lineSplit[0].equals("EFIX")) {

					//the start and end timestamp of this event
					long startStamp = Long.parseLong(lineSplit[2]);
					long endStamp = Long.parseLong(lineSplit[3]);

					//the pupilsize of this event
					float pupilSize = Float.parseFloat(lineSplit[lineSplit.length - 1]);

					//although, this value is the area of the pupil. we want radius
					pupilSize = (float) Math.sqrt(pupilSize / Math.PI);

					//quick sanity check...
					assert !dataPacketsL.isEmpty() && !dataPacketsR.isEmpty();

					//the vector we're inserting into
					Vector whichVector = null;
					if (lineSplit[1].equals("L")) {
						//sanity checks...
						assert ((DataPacket) dataPacketsL.lastElement())
							.getTimeStamp() <= endStamp : ((PupilDataPacket) dataPacketsL
							.lastElement()).getTimeStamp()
							+ "<=" + endStamp;

						whichVector = dataPacketsL;
					} else {
						//sanity checks...
						assert ((DataPacket) dataPacketsR.lastElement())
							.getTimeStamp() <= endStamp : ((PupilDataPacket) dataPacketsR
							.lastElement()).getTimeStamp()
							+ "<=" + endStamp;
	
						whichVector = dataPacketsR;
					}

					for (long time = startStamp; time < endStamp; time += 10) {
						whichVector.add(new PupilDataPacket(time, pupilSize));
					}
					whichVector.add(new PupilDataPacket(endStamp, pupilSize));

					
				}
			}

		} else {
			// only get pupil information during a fixation
			boolean leftFixate = false;
			long prevLeftTimeStamp = 0;
			float prevLeftPupilSize = 0;

			boolean rightFixate = false;
			long prevRightTimeStamp = 0;
			float prevRightPupilSize = 0;

			while ((tmp = inputStream.readLine()) != null) {

				lineSplit = tmp.split("\\s+");

				//start event
				if (lineSplit[0].equals("START")) {

					//the timestamp of this event
					long timeStamp = Long.parseLong(lineSplit[1]);

					trialStartL.add(new Integer(dataPacketsL.size()));
					dataPacketsL.add(new StartPacket(timeStamp, trialStartL.size()));

					trialStartR.add(new Integer(dataPacketsR.size()));
					dataPacketsR.add(new StartPacket(timeStamp, trialStartR.size()));

				} else if (lineSplit[0].equals("END")) {
					//the timestamp of this event
					long timeStamp = Long.parseLong(lineSplit[1]);

					dataPacketsL.add(new EndPacket(timeStamp, trialStartL.size()));
					dataPacketsR.add(new EndPacket(timeStamp, trialStartR.size()));


				} else if (lineSplit[0].equals("SFIX")) {
					if (lineSplit[1].equals("L")) {
						leftFixate = true;
					} else if (lineSplit[1].equals("R")) {
						rightFixate= true;
					}
	

				} else if (lineSplit[0].equals("EFIX")) {

					if (lineSplit[1].equals("L")) {
						leftFixate = false;
					} else if (lineSplit[1].equals("R")) {
						rightFixate= false;
					}

				} else if ( (lineSplit[0].length() > 0) && (Character.isDigit(lineSplit[0].charAt(0)))) {

					long timeStamp = Long.parseLong(lineSplit[0]);

					//the pupilsize of this event
					float leftPupilSize = Float.parseFloat(lineSplit[3]);
					float rightPupilSize = Float.parseFloat(lineSplit[6]);

					//although, this value is the area of the pupil. we want radius
					leftPupilSize = (float) Math.sqrt(leftPupilSize / Math.PI);
					rightPupilSize = (float) Math.sqrt(rightPupilSize / Math.PI);

					if (leftFixate) {
						// There may have been a time gap due to not including data within saccades and other events
						float pIncrement = TIME_INCREMENT * ((leftPupilSize - prevLeftPupilSize) / (timeStamp - prevLeftTimeStamp));
						for (long t=prevLeftTimeStamp + TIME_INCREMENT; t<timeStamp; t += TIME_INCREMENT) 
						{
							dataPacketsL.add(new PupilDataPacket(t, prevLeftPupilSize + pIncrement));
							prevLeftPupilSize += pIncrement;
						}

						if (leftPupilSize > MINIMUM_VALID_SIZE) 
						{
							dataPacketsL.add(new PupilDataPacket(timeStamp, leftPupilSize));
						} else {
							dataPacketsL.add(new PupilDataPacket(timeStamp, 0));
						}

						prevLeftTimeStamp = timeStamp;
						prevLeftPupilSize = leftPupilSize;
					}

					if (rightFixate) {

						// There may have been a time gap due to not including data within saccades and other events
						float pIncrement = TIME_INCREMENT * ((rightPupilSize - prevRightPupilSize) / (timeStamp - prevRightTimeStamp));
						for (long t=prevRightTimeStamp + TIME_INCREMENT; t<timeStamp; t += TIME_INCREMENT) 
						{
							dataPacketsR.add(new PupilDataPacket(t, prevRightPupilSize + pIncrement));
							prevRightPupilSize += pIncrement;
						}


						if (rightPupilSize > MINIMUM_VALID_SIZE) 
						{
							dataPacketsR.add(new PupilDataPacket(timeStamp, rightPupilSize));
						} 
						else 
						{
							dataPacketsR.add(new PupilDataPacket(timeStamp, 0));
						}


						prevRightTimeStamp = timeStamp;
						prevRightPupilSize = rightPupilSize;

					}
				}
			}

		}

		inputStream.close();

		// OK, now lets find and replace the blinks (which have 0 values)
		dataPacketsL = removeBlinks(dataPacketsL);

		dataPacketsR = removeBlinks(dataPacketsR);


	}




	private Vector removeBlinks(Vector dataPackets) {

		Vector newDataPackets = new Vector();

		for (int i=0; i<dataPackets.size(); ++i) {

			if (dataPackets.elementAt(i) instanceof StartPacket) {
				StartPacket sp = (StartPacket) dataPackets.elementAt(i);
				newDataPackets.add(new StartPacket(sp.getTimeStamp(), sp.getTrial()));
				continue;
			}


			if (dataPackets.elementAt(i) instanceof EndPacket) {
				EndPacket ep = (EndPacket) dataPackets.elementAt(i);
				newDataPackets.add(new EndPacket(ep.getTimeStamp(), ep.getTrial()));
				continue;
			}


			PupilDataPacket p = (PupilDataPacket) dataPackets.elementAt(i);

			float pupilSize = p.getPupilSize();

			if (pupilSize == 0.0) {

				int j;
				for (j=i+1; ((j<dataPackets.size()) && (((PupilDataPacket) dataPackets.elementAt(j)).getPupilSize() == 0.0)); ++j) {
				}

				// still in range
				if (j < dataPackets.size()) {
					float prevValue = getPrevValue(newDataPackets, i-1, 10); // use new vector to get prevValue
					float increment = ( (((PupilDataPacket) dataPackets.elementAt(j)).getPupilSize() - prevValue) / (j-i+1));

				   	// interpolate
				   	for (int a=i; a<j; ++a) {
						PupilDataPacket pp = (PupilDataPacket) dataPackets.elementAt(a);
						newDataPackets.add(new PupilDataPacket(pp.getTimeStamp(), (prevValue + increment)));
						prevValue += increment;
					}

					newDataPackets.add(new PupilDataPacket(
						((PupilDataPacket) dataPackets.elementAt(j)).getTimeStamp(),
						((PupilDataPacket) dataPackets.elementAt(j)).getPupilSize() ));
				}

				i=j;

			} else {
				newDataPackets.add(new PupilDataPacket(p.getTimeStamp(), pupilSize));
			
			}

		}

		return newDataPackets;

	}






	// Get an average of the last Window values
	private float getPrevValue(Vector v, int start, int window) {
		float sum = 0;
		int j=0;
		for (j=start; ((j>=0) && (j>start-window)); --j) 
		{
			if (v.elementAt(j) instanceof PupilDataPacket) 
			{	
				sum += ((PupilDataPacket) v.elementAt(j)).getPupilSize();
			} 
		}
		return (sum / (start - j + 1));
	}




	public static Vector meanFilter(Vector dataPackets, int filter_window_size) 
	{
		Vector newDataPackets = new Vector();
		int halfSize = filter_window_size / 2;

		for (int i=halfSize + 1; i<dataPackets.size() - (halfSize + 1); ++i) 
		{
			if (dataPackets.elementAt(i) instanceof StartPacket) 
			{
				StartPacket sp = (StartPacket) dataPackets.elementAt(i);
				newDataPackets.add(new StartPacket(sp.getTimeStamp(), sp.getTrial()));
				continue;
			}

			if (dataPackets.elementAt(i) instanceof EndPacket) 
			{
				EndPacket ep = (EndPacket) dataPackets.elementAt(i);
				newDataPackets.add(new EndPacket(ep.getTimeStamp(), ep.getTrial()));
				continue;
			}

			PupilDataPacket p = (PupilDataPacket) dataPackets.elementAt(i);

			float avg = 0;
			int count = 0;
			for (int j=0; j<filter_window_size; ++j) 
			{
				float val = getValueAt(dataPackets, j + i - halfSize);
				//System.out.println(val);
				if (val > 0) 
				{
					avg += val;
					++count;
				}
			}
			if (count > 0) 
			{
				avg /= count;
			}
			//System.out.println(avg);
			//System.exit(1);
			newDataPackets.add(new PupilDataPacket(p.getTimeStamp(), avg));
		}

		return newDataPackets;

	}

	
	public static float getValueAt(Vector v, int i) 
	{
		if ((i <= 0) || (v.elementAt(i) instanceof StartPacket) || (v.elementAt(i) instanceof EndPacket)) 
		{
			return 0;
		}
		return (((PupilDataPacket)v.elementAt(i)).getPupilSize());

	}




	// written by BPB - a quick hack to determine if the file has sample data in it (as opposed to only events)

	private boolean hasSampleData() throws IOException {
		
		BufferedReader stream = null;

		try {
			stream = new BufferedReader(new FileReader(file));
		} catch (Exception ex) {
			System.out.println(file + " not found!");
			System.exit(1);
		}


		String tmp;
		String[] lineSplit;
		boolean hasSamples = false;
		boolean done = false;		

		while (! done) {
			tmp = stream.readLine();
			if (tmp != null) {
				lineSplit = tmp.split("\\s+");

				// start event
				if (lineSplit[0].equals("START")) {
					String tmp2;
					String[] lineSplit2;

					while (! done) {
						tmp2 = stream.readLine();
						if (tmp2 != null) {
							lineSplit2 = tmp2.split("\\s+");
							if (lineSplit2[0].equals("EVENTS")) {
								String tmp3 = stream.readLine();
								String[] lineSplit3 = tmp3.split("\\s+");
								if (lineSplit3[0].equals("SAMPLES")) {
									hasSamples = true;
									done = true;
								} else {
									done = true;									
								}

							}
								
						} else {
							done = true;
						}
					}
				}

			} else {
				done = true;
			}
		}

		stream.close();
		return hasSamples;

	}

			



	/**
	 * this must be called after the file is parsed
	 * 
	 * @param minTime -
	 *            the min time of the baseline trial boundary
	 * @param maxTime -
	 *            the max time of the baselien trial boundary
	 * @param eye -
	 *            ReadPupil.LEFT, ReadPupil.RIGHT
	 * @return
	 */
	public Vector getData(long minTime, long maxTime, int eye) {
		//catch error
		if (minTime > maxTime)
			return null;

		//get the data packets we are interested in
		Vector dataPacketsSrc;
		if (eye == LEFT) {
			dataPacketsSrc = dataPacketsL;
		} else {
			dataPacketsSrc = dataPacketsR;
		}

		//again, try to catch error
		if (minTime > ((DataPacket) dataPacketsSrc.lastElement())
				.getTimeStamp()
				|| maxTime < ((DataPacket) dataPacketsSrc.firstElement())
						.getTimeStamp())
			return null;

		//find the start and end indices
		int startindex = 0;
		int endindex = 0;
		for (int i = 0; i < dataPacketsSrc.size(); i++) {
			if (((DataPacket) dataPacketsSrc.get(i)).getTimeStamp() > minTime) {
				startindex = i;
				break;
			}
		}

		endindex = startindex;
		for (int i = startindex + 1; i < dataPacketsSrc.size(); i++) {
			if (((DataPacket) dataPacketsSrc.get(i)).getTimeStamp() > maxTime) {
				break;
			}
			endindex = i;
		}

		//finally, return the relevant data packets
		return new Vector(dataPacketsL.subList(startindex, endindex));
	}

	/**
	 * this must be called after the file is parsed
	 * 
	 * @param trial -
	 *            ReadPupil.ALL_TRIALS or a trial number in range of 0 -
	 *            (getNumTrials()-1)
	 * @param eye -
	 *            ReadPupil.LEFT, ReadPupil.RIGHT
	 * @return - the data packets for these parameters
	 */
	public Vector getData(int trial, int eye) {

		//returning all trials
		if (trial == ALL_TRIALS) {
			if (eye == LEFT)
				return dataPacketsL;
			else
				return dataPacketsR;
		}

		//catch error
		if (trial < 0 || trial >= getNumTrials())
			return null;

		Vector dataPacketsSrc;
		Vector trialStartSrc;

		if (eye == LEFT) {
			dataPacketsSrc = dataPacketsL;
			trialStartSrc = trialStartL;
		} else {
			dataPacketsSrc = dataPacketsR;
			trialStartSrc = trialStartR;
		}

		int startindex = ((Integer) trialStartSrc.get(trial)).intValue();
		int endindex;

		if (trial == getNumTrials() - 1)
			endindex = dataPacketsL.size();
		else
			endindex = ((Integer) trialStartSrc.get(trial + 1)).intValue();

		return new Vector(dataPacketsL.subList(startindex, endindex));
	}

	/**
	 * @return Returns the number of trials read from file.
	 */
	public int getNumTrials() {
		if (trialStartL == null || trialStartR == null)
			return 0;

		//the size of trialStartL and R should be the same...
		assert trialStartL.size() == trialStartR.size();

		return trialStartL.size();
	}

	/**
	 * method returns a 2D array of longs. the length of the array is the number
	 * of trials. each trial cell contains a length 2 array of start and end
	 * times
	 * 
	 * @return
	 */
	public long[][] getTrialData() {
		long trialData[][] = new long[getNumTrials()][];

		for (int i = 0; i < trialStartL.size(); i++) {
			trialData[i] = new long[2];

			trialData[i][0] = ((DataPacket) dataPacketsL
					.get(((Integer) trialStartL.get(i)).intValue()))
					.getTimeStamp();

			if (i == trialStartL.size() - 1)
				trialData[i][1] = ((DataPacket) dataPacketsL.lastElement())
						.getTimeStamp();
			else
				trialData[i][1] = ((DataPacket) dataPacketsL
						.get(((Integer) trialStartL.get(i + 1)).intValue() - 1))
						.getTimeStamp();
		}

		return trialData;
	}

	/**
	 * @param Trial -
	 *            Trial data returned from ReadPupil.getData
	 */
	public static Vector prunePupilData(Vector testTrial) {
		Vector pupilData = new Vector();

		for (Iterator iter = testTrial.iterator(); iter.hasNext();) {
			Object element = (Object) iter.next();

			if (element instanceof PupilDataPacket) {
				PupilDataPacket PD = (PupilDataPacket) element;
				pupilData.add(PD);
			}

		}

		return pupilData;
	}

	/**
	 * @param baseLineTrial -
	 *            baseline vector returned from ReadPupil.getData
	 */
	public static float calculateBaseLine(Vector baseLineTrial) {
		float basePupil = 0;
		int numCounted = 0;
		for (Iterator iter = baseLineTrial.iterator(); iter.hasNext();) {
			Object element = (Object) iter.next();

			if (element instanceof PupilDataPacket) {
				PupilDataPacket PD = (PupilDataPacket) element;
				basePupil += PD.getPupilSize();
				numCounted++;
			}
		}
		if (numCounted > 0)
			basePupil /= numCounted;

		return basePupil;
	}

	/**
	 * 
	 * @param trialData
	 * @param baseLine
	 * @return
	 */
	public static float calculateMaxAbsPCPS(Vector trialData, float baseLine) {
		float maxAbsPCPS = -1;
		for (Iterator iter = trialData.iterator(); iter.hasNext();) {
			Object element = (Object) iter.next();

			if (element instanceof PupilDataPacket) {
				PupilDataPacket PD = (PupilDataPacket) element;

				if ((maxAbsPCPS == -1
						|| maxAbsPCPS < Math.abs(PD.getPupilSize() - baseLine))) {
					maxAbsPCPS = Math.abs(PD.getPupilSize() - baseLine);
				}
			}
		}

		return maxAbsPCPS / baseLine;
	}

	/**
	 * method calculates the max pcps
	 * 
	 * @param trialData
	 * @param baseLine
	 * @return
	 */
	public static float calculateMaxPCPS(Vector trialData, float baseLine) {
		float maxPCPS = -1;
		boolean maxDefined = false;

		for (Iterator iter = trialData.iterator(); iter.hasNext();) {
			Object element = (Object) iter.next();

			if (element instanceof PupilDataPacket) {
				PupilDataPacket PD = (PupilDataPacket) element;

				if ((!maxDefined || maxPCPS < PD.getPupilSize() - baseLine)) {
					maxDefined = true;
					maxPCPS = PD.getPupilSize() - baseLine;
				}
			}
		}

		return maxPCPS / baseLine;
	}

	/**
	 * method calculates the min pcps
	 * 
	 * @param trialData
	 * @param baseLine
	 * @return
	 */
	public static float calculateMinPCPS(Vector trialData, float baseLine) {
		float minPCPS = -1;
		boolean minDefined = false;

		for (Iterator iter = trialData.iterator(); iter.hasNext();) {
			Object element = (Object) iter.next();

			if (element instanceof PupilDataPacket) {
				PupilDataPacket PD = (PupilDataPacket) element;

				if ((PD.getPupilSize() > MINIMUM_VALID_SIZE) && (!minDefined || minPCPS > PD.getPupilSize() - baseLine)) {
					minDefined = true;
					minPCPS = PD.getPupilSize() - baseLine;
				}
			}
		}

		return minPCPS / baseLine;
	}

	/**
	 * method returns the minimum timestamp for a given eye
	 * 
	 * @param eye
	 *            ReadPupil.LEFT, ReadPupil.RIGHT
	 * @return the min time
	 */
	public long getMin(int eye) {
		Vector packetsSrc = null;
		if (eye == ReadPupil.LEFT) {
			packetsSrc = dataPacketsL;
		} else {
			packetsSrc = dataPacketsR;
		}

		return ((DataPacket) packetsSrc.firstElement()).getTimeStamp();
	}

	/**
	 * method returns the maximum timestamp for a given eye
	 * 
	 * @param eye
	 *            ReadPupil.LEFT, ReadPupil.RIGHT
	 * @return the max time
	 */
	public long getMax(int eye) {
		Vector packetsSrc = null;
		if (eye == ReadPupil.LEFT) {
			packetsSrc = dataPacketsL;
		} else {
			packetsSrc = dataPacketsR;
		}

		return ((DataPacket) packetsSrc.lastElement()).getTimeStamp();
	}

}