/*
 * Created on Sep 16, 2004
 * TaskView.java
 * 
 * Project: eyetrack_cvs
 */
package task;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextLayout;
import java.awt.image.BufferedImage;
import java.text.AttributedString;
import java.text.DecimalFormat;
import java.util.Vector;

import view.VisWindow;

/**
 * @author chris
 *  
 */
public class TaskView extends VisWindow {

	/**
	 * the taskBlocks to draw
	 */
	private Vector taskBlocks = null;

	/**
	 * the level colors
	 */
	private Color[] levelColors = null;

	/** the buffered image for this view* */
	private BufferedImage viewBuffer = null;

	/** the min buffer time */
	private long minBufTime = 0;

	/** the max buffer time */
	private long maxBufTime = 0;

	/** the min current time* */
	private long minViewTime = 0;

	/** the max current time* */
	private long maxViewTime = 0;

	protected Rectangle drawRegion;

	/**
	 * default constructor. takes as input the drawing region.
	 * 
	 * @param drawRegion
	 */
	public TaskView(Rectangle drawRegion) {
		this.drawRegion = drawRegion;
	}

	/**
	 * method that actually draws the window
	 * 
	 * @param g2
	 */
	public void drawWindow(Graphics2D g2) {
		if (taskBlocks == null)
			return;

		//initialize the buffered region if we have to
		if (viewBuffer == null || viewBuffer.getWidth() != drawRegion.width
				|| viewBuffer.getHeight() != drawRegion.height) {

			//initialize the buffer...
			viewBuffer = new BufferedImage(drawRegion.width, drawRegion.height,
					BufferedImage.TYPE_INT_RGB);

			//update the times...
			minBufTime = minViewTime;
			maxBufTime = maxViewTime;

			Graphics2D tmpG = viewBuffer.createGraphics();
			Rectangle tmp = new Rectangle(0, 0, viewBuffer.getWidth(),
					viewBuffer.getHeight());
			tmpG.setColor(Color.WHITE);
			tmpG.fill(tmp);
			drawTaskModel(tmpG);
		}

		//check if the buffered image has expired
		if (minBufTime != minViewTime || maxBufTime != maxViewTime) {
			//reset the times...
			minBufTime = minViewTime;
			maxBufTime = maxViewTime;

			Graphics2D tmpG = viewBuffer.createGraphics();
			Rectangle tmp = new Rectangle(0, 0, viewBuffer.getWidth(),
					viewBuffer.getHeight());
			tmpG.setColor(Color.WHITE);
			tmpG.fill(tmp);
			drawTaskModel(tmpG);
		}

		//draw into the focus view
		g2.drawImage(viewBuffer, (int) drawRegion.getMinX(), (int) drawRegion
				.getMinY(), null);
	}

	/**
	 * method draws the task model- should be called on buffered image
	 * 
	 * @param g2
	 */
	private void drawTaskModel(Graphics2D g2) {
		//the height of each task block
		int height = (int) (drawRegion.getHeight() / taskBlocks.size());
		//the remainder pixels
		int remainder = (int) drawRegion.getHeight() - height
				* taskBlocks.size();

		//the top pixel of the level
		int top = 0;
		//the bottom pixel of the level
		int bottom = 0;

		Rectangle clipRect = new Rectangle(drawRegion);
		clipRect.setLocation(0, 0);

		//some text stuff...
		FontRenderContext frc = g2.getFontRenderContext();

		//draw each level
		for (int level = 0; level < taskBlocks.size(); level++) {
			Vector blocks = (Vector) taskBlocks.get(level);

			//set the bottom pixel
			if (remainder > 0) {
				bottom = top + height + 1;
				remainder--;
			} else
				bottom = top + height;

			for (int i = 0; i < blocks.size(); i++) {
				//check if we even need to draw this task block
				TaskBlock tb = (TaskBlock) blocks.get(i);
				if (tb.getEndTime() < minViewTime)
					continue;
				if (tb.getStartTime() > maxViewTime)
					break;

				//the startx, endx
				int startX = (int) getXCoord(tb.getStartTime(), clipRect);
				int endX = (int) getXCoord(tb.getEndTime(), clipRect);

				//the drawn block
				Rectangle tbRect = new Rectangle((int) startX, (int) top,
						(int) (endX - startX), bottom - top);

				g2.setColor(levelColors[level % levelColors.length]);
				g2.fill(tbRect);

				//draw outline...
				g2.setColor(Color.BLACK);
				g2.setStroke(new BasicStroke(.2f));
				g2.draw(tbRect);

				//try to draw the taskBlock name
				g2.setColor(Color.BLACK);
				AttributedString theName = new AttributedString(tb.getName());
				LineBreakMeasurer lbm = new LineBreakMeasurer(theName
						.getIterator(), frc);
				TextLayout tl;

				//if this task block is "cut off" by the view, reposition the
				// text starting point
				int startText = (int) tbRect.getMinX() + 2;
				if (tb.getStartTime() < minViewTime) {
					startText = 2;
				}

				try {
					tl = lbm.nextLayout((float) tbRect.getMaxX() - startText,
							tb.getName().length(), true);
					if (tl != null) {
						tl.draw(g2, startText, (int) tbRect.getMaxY() - 2);
						//see if dots are necessary
						if (lbm.getPosition() < tb.getName().length()) {
							theName = new AttributedString("...");
							lbm = new LineBreakMeasurer(theName.getIterator(),
									frc);
							int width = (int) tl.getBounds().getWidth();
							tl = lbm.nextLayout((float) (tbRect.getWidth()
									- width - 1), 3, true);
							if (tl != null)
								tl.draw(g2, (int) (startText + width + 1),
										(int) tbRect.getMaxY() - 2);
						}
					} else {
						//see if dots will fit...
						theName = new AttributedString("...");
						lbm = new LineBreakMeasurer(theName.getIterator(), frc);

						tl = lbm.nextLayout((float) tbRect.getWidth() - 2, 3,
								true);
						if (tl != null)
							tl.draw(g2, (int) tbRect.getMinX() + 2,
									(int) tbRect.getMaxY() - 2);
					}
				} catch (Exception e) {
				}
			}

			//set the top...
			top = bottom;
		}
	}

	/**
	 * given some time, method returns the x coords of the time
	 * 
	 * @param time -
	 *            the time in question
	 */
	public int getXCoord(double time) {
		return getXCoord(time, drawRegion);
	}

	/**
	 * given some rectangle of interest, method returns the x coord at that time
	 * 
	 * @param time
	 * @param r
	 *            most likely the cliprect used for drawing to the buffered
	 *            region
	 * @return
	 */
	private int getXCoord(double time, Rectangle r) {
		if (minViewTime == maxViewTime)
			return -1;
		return (int) (r.getMinX() + ((r.getWidth() - 1) * (time - minViewTime) / (maxViewTime - minViewTime)));
	}

	/**
	 * method sets the data for this view
	 * 
	 * @param taskBlocks
	 * @param levelColors
	 */
	public void setData(Vector taskBlocks, Color[] levelColors) {
		this.taskBlocks = taskBlocks;
		this.levelColors = levelColors;
	}

	/**
	 * method sets the current viewing times
	 * 
	 * @param minViewTime
	 * @param maxViewTime
	 */
	public void setViewTimes(long minViewTime, long maxViewTime) {
		this.minViewTime = minViewTime;
		this.maxViewTime = maxViewTime;
	}

	/**
	 * given some xposition, method returns the time at that point
	 * 
	 * @param xPos
	 * @return
	 */
	public double getTime(int xPos) {
		if (drawRegion.width == 0)
			return -1;
		return minViewTime + (maxViewTime - minViewTime)
				* (xPos - drawRegion.getMinX()) / (drawRegion.width);
	}

	/**
	 * method draws the labels for this view
	 * 
	 * @param g2
	 * @param f
	 *            the font to use
	 */
	public void drawLabels(Graphics2D g2, Font f) {
		if (taskBlocks != null) {
			g2.setColor(Color.GRAY);
			FontRenderContext frc = g2.getFontRenderContext();

			//draw label for each level
			//the height of each task block
			int height = (int) (drawRegion.getHeight() / taskBlocks.size());
			//the remainder pixels
			int remainder = (int) drawRegion.getHeight() - height
					* taskBlocks.size();

			//the top pixel of the level
			int top = drawRegion.y;
			//the bottom pixel of the level
			int bottom = 0;

			//the x position should be consistent...
			int xPos = -1;
			for (int level = 0; level < taskBlocks.size(); level++) {

				//set the bottom pixel
				if (remainder > 0) {
					bottom = top + height + 1;
					remainder--;
				} else
					bottom = top + height;

				TextLayout tl = new TextLayout("lv " + level, f, frc);
				if (xPos == -1)
					xPos = (int) (drawRegion.getMinX()
							- tl.getBounds().getWidth() - 5);

				if (tl.getBounds().getHeight() < height)
					tl.draw(g2, xPos, bottom
							- (int) (height - tl.getBounds().getHeight()) / 2);
				else
					tl.draw(g2, xPos, bottom);

				top = bottom;
			}
		}
	}

	public String getToolTip(java.awt.event.MouseEvent e) {
		//the level this event occurs at
		int level = -1;

		//the "time" where the mouse was pressed
		double mouseTime = getTime(e.getX());

		//the height of the blocks, and the remainder
		int height = (int) (drawRegion.getHeight() / taskBlocks.size());
		int remainder = (int) drawRegion.getHeight() - height
				* taskBlocks.size();

		//the bottom of level...
		int bottom = (int) drawRegion.getMinY();
		for (level = 0; level < taskBlocks.size(); level++) {
			//get the bottom of this level...
			if (remainder > 0) {
				bottom += height + 1;
				remainder--;
			} else
				bottom += height;

			//found the level?
			if (e.getY() <= bottom)
				break;
		}

		//nothing to find here...
		if (level >= taskBlocks.size())
			return null;

		//get the vector for this level, find the block
		Vector levelVec = (Vector) taskBlocks.get(level);
		TaskBlock theBlock = null;
		for (int i = 0; i < levelVec.size(); i++) {
			TaskBlock tb = (TaskBlock) levelVec.get(i);
			if (mouseTime < tb.getStartTime())
				break;
			if (mouseTime >= tb.getStartTime() && mouseTime <= tb.getEndTime()) {
				theBlock = tb;
				break;
			}
		}

		//nothing found?
		if (theBlock == null)
			return null;

		DecimalFormat df = new DecimalFormat("#.##");

		if (theBlock != null) {
			return "<html>" + theBlock.getName() + "<br>Level " + (level + 1)
					+ " Task" + "<br>" + df.format(theBlock.getStartTime())
					+ " - " + df.format(theBlock.getEndTime());
		}

		//nothing...
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getMinViewTime()
	 */
	public long getMinViewTime() {
		return minViewTime;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getMaxViewTime()
	 */
	public long getMaxViewTime() {
		return maxViewTime;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#setDrawRegion(java.awt.Rectangle)
	 */
	public void setDrawRegion(Rectangle drawRegion) {
		this.drawRegion = drawRegion;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getDrawRegion()
	 */
	public Rectangle getDrawRegion() {
		return drawRegion;
	}
}