/*
 * Created on Sep 22, 2004
 * TimeRuler.java
 * 
 * Project: eyetrack_cvs
 */
package view;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.image.BufferedImage;
import java.text.DecimalFormat;

/**
 * @author chris
 *  
 */
public class TimeRuler extends VisWindow {

	/** the buffered image for this ruler* */
	private BufferedImage viewBuffer = null;

	/** the min time of the buffered image* */
	private long minBufTime = 0;

	/** the max time of the buffered image* */
	private long maxBufTime = 0;

	/** the current min time of the time ruler* */
	private long minTime = 0;

	/** the current max time of the time ruler* */
	private long maxTime = 0;

	/** the draw region...* */
	private Rectangle drawRegion;

	/** the minimum pixels per minor tick* */
	private int minPixPerMinTick = 5;

	/** the minimum pixels per major tick* */
	private int minPixPerMajTick;

	/** the minor tick height* */
	private int minorHeight = 3;

	/** the major tick height* */
	private int majorHeight = 9;

	/** if flipped is true, the ticks are positioned on top* */
	private boolean flipped = true;

	/** number formatter */
	private DecimalFormat df = new DecimalFormat("#");

	/**
	 * default constructor
	 * 
	 * @param drawRegion -
	 *            the rectangle to draw to
	 * @param flipped -
	 *            if flipped is true, the ticks are posiitoned on false
	 */
	public TimeRuler(Rectangle drawRegion, boolean flipped) {
		this.drawRegion = drawRegion;
		this.flipped = flipped;
		minTime = maxTime = -1;
	}

	/**
	 * method sets the boundary time of the ruler
	 * 
	 * @param minTime
	 * @param maxTime
	 */
	public void setViewTimes(long minTime, long maxTime) {
		this.minTime = minTime;
		this.maxTime = maxTime;
	}

	/**
	 * method draws time ruler onto a Graphics2D object
	 * 
	 * @param g2
	 */
	public void drawWindow(Graphics2D g2) {

		//only draw if we have a range of times
		if (minTime == maxTime)
			return;

		//initialize the buffered region if we have to
		if (viewBuffer == null || viewBuffer.getWidth() != drawRegion.width
				|| viewBuffer.getHeight() != drawRegion.height) {

			//initialize the focusBuffer...
			viewBuffer = new BufferedImage(drawRegion.width, drawRegion.height,
					BufferedImage.TYPE_INT_RGB);

			//update the times...
			minBufTime = minTime;
			maxBufTime = maxTime;

			Graphics2D tmpG = viewBuffer.createGraphics();
			Rectangle tmp = new Rectangle(0, 0, viewBuffer.getWidth(),
					viewBuffer.getHeight());
			tmpG.setColor(Color.WHITE);
			tmpG.fill(tmp);
			drawRuler(tmpG);
		}

		//check if the buffered image has expired
		if (minBufTime != minTime || maxBufTime != maxTime) {
			//reset the times...
			minBufTime = minTime;
			maxBufTime = maxTime;

			Graphics2D tmpG = viewBuffer.createGraphics();
			Rectangle tmp = new Rectangle(0, 0, viewBuffer.getWidth(),
					viewBuffer.getHeight());
			tmpG.setColor(Color.WHITE);
			tmpG.fill(tmp);
			drawRuler(tmpG);
		}

		//actually draw the buffered image to the screen
		g2.drawImage(viewBuffer, (int) drawRegion.getMinX(), (int) drawRegion
				.getMinY(), null);
	}

	/**
	 * method should be called to draw onto the buffered image
	 * 
	 * @param g2
	 */
	private void drawRuler(Graphics2D g2) {
		//an estimate of the minor spacing
		int spacing = minPixPerMinTick;

		//compute the ms per min tick, adjust the spacing.
		double msPerMinTick = 5. * (maxTime - minTime) / drawRegion.getWidth();
		msPerMinTick = Math.ceil(msPerMinTick);

		if (msPerMinTick > 3 && msPerMinTick % 5 != 0)
			msPerMinTick += 5 - msPerMinTick % 5;

		spacing = (int) ((msPerMinTick) * drawRegion.getWidth() / (maxTime - minTime));

		//get the major tick spacing requirement based on the bounds
		//of the max time label
		g2.setColor(Color.BLACK);
		g2.setStroke(new BasicStroke(1.0f));

		Font f = new Font("Helvetica", Font.PLAIN, 10);
		TextLayout tl;
		FontRenderContext frc = g2.getFontRenderContext();

		tl = new TextLayout(Double.toString(maxTime), f, frc);
		minPixPerMajTick = (int) (1.5 * tl.getBounds().getWidth());

		int minTickPerMajTick = (int) Math.ceil((double) minPixPerMajTick
				/ (double) spacing);
		if (minTickPerMajTick > 2 && minTickPerMajTick % 5 != 0)
			minTickPerMajTick += 5 - minTickPerMajTick % 5;

		//the current tick we're looking at
		double theTick = minTime + msPerMinTick - (minTime % msPerMinTick);

		//the x location of this tick
		//int x=drawRegion.x+(int)(spacing*((theTick-minTime)/msPerMinTick));
		int x = (int) (spacing * ((theTick - minTime) / msPerMinTick));

		//the number of ticks we've gotten so far
		int count = 0;

		//the remainder we need to make up
		int remainder = (int) (drawRegion.width - x - spacing
				* (int) Math.ceil((maxTime - minTime) / msPerMinTick));
		if (remainder < 0)
			remainder = 0;

		//draw the ticks...
		for (; theTick < maxTime; theTick += msPerMinTick) {
			if (remainder != 0) {
				if ((count
						% (int) ((maxTime - minTime) / (msPerMinTick * remainder)) == 0)) {
					x++;
				}
			}

			if (theTick % (minTickPerMajTick * msPerMinTick) == 0) {
				tl = new TextLayout(df.format(theTick), f, frc);

				if (flipped) {
					g2.drawLine(x, 0, x, drawRegion.height);
					tl.draw(g2, x + 1, minorHeight
							+ (int) tl.getBounds().getHeight() + 3);
				} else {
					g2.drawLine(x, drawRegion.height, x, 0);
					tl.draw(g2, x + 1, drawRegion.height - minorHeight - 2);
				}
			} else {
				if (flipped) {
					g2.drawLine(x, 0, x, minorHeight);
				} else {
					g2.drawLine(x, drawRegion.height, x, drawRegion.height
							- minorHeight);
				}
			}

			//move x forward
			x += spacing;

			//increment count
			count++;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getMinViewTime()
	 */
	public long getMinViewTime() {
		return minTime;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getMaxViewTime()
	 */
	public long getMaxViewTime() {
		return maxTime;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#setDrawRegion(java.awt.Rectangle)
	 */
	public void setDrawRegion(Rectangle drawRegion) {
		this.drawRegion = drawRegion;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getDrawRegion()
	 */
	public Rectangle getDrawRegion() {
		return drawRegion;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getTime(int)
	 */
	public double getTime(int x) {
		if (drawRegion.width == 0)
			return -1;
		return minTime + (maxTime - minTime) * (x - drawRegion.getMinX())
				/ (drawRegion.width);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see view.VisWindow#getXCoord(double)
	 */
	public int getXCoord(double time) {
		if (minTime == maxTime)
			return -1;
		return (int) (drawRegion.getMinX() + ((drawRegion.getWidth() - 1)
				* (time - minTime) / (maxTime - minTime)));
	}
}

